/*
 * ThreadPool.h - by Alan Alvarez <alan.alvarez@us.army.mil>
 *
 * Mira GroupWare
 * Copyright (C) 2007 Mira GroupWare Dev Team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __MIRACLIENT_THREADPOOL_H
#define __MIRACLIENT_THREADPOOL_H

#include <boost/thread.hpp>
#include <boost/thread/condition.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <queue>

#include <iostream>

namespace miraclient
{

class ThreadPool
{
//    typedef task_type TaskType;

    public:
        ThreadPool(const unsigned int thread_count) : m_thread_count(thread_count), m_used_threads_count(0), m_task_queue()
        {
            for (unsigned int i = 0; i < m_thread_count; ++i)
                boost::thread(boost::bind(&ThreadPool::run, this));
        }

        ThreadPool(const ThreadPool& rhs)
        {
            // This shouldn't really be used.
            m_thread_count = 0;
            m_used_threads_count = 0;
        }

        ThreadPool() : m_thread_count(0), m_used_threads_count(0), m_task_queue()
        {}

        unsigned int size() const { return m_thread_count; }
 
        void add_thread(const unsigned int threads_to_add = 1)
        {
            m_thread_count += threads_to_add;
            for (unsigned int i = 0; i < threads_to_add; ++i)
                boost::thread(boost::bind(&ThreadPool::run, this));
        }

        void add_task(const boost::function0<void>& task)
        {
            if (!m_thread_count)
                add_thread(); // Make sure there is at least one thread running.

            boost::mutex::scoped_lock queue_lock(m_queue_mutex);
            m_task_queue.push(task);
            queue_lock.unlock();
            
            if (m_used_threads_count != m_thread_count)
                m_new_task_condition.notify_one();
        }

        void run()
        {
            do
            {
                boost::mutex::scoped_lock lock(m_new_task_mutex);
                m_new_task_condition.wait(lock);
                lock.unlock();
                
                ++m_used_threads_count;
                while (true)
                {   
                    // Keep running tasks until there aren't any more tasks scheduled.
                    boost::mutex::scoped_lock queue_lock(m_queue_mutex);
                    boost::function0<void> function_copy;
                    if (!m_task_queue.empty())
                    {
                        function_copy = m_task_queue.front();
                        m_task_queue.pop();
                        queue_lock.unlock();
                    }
                    else
                    {
                        queue_lock.unlock();
                        break;
                    }

                    function_copy();
                }
                --m_used_threads_count;
            }
            while (true);
        }

    private:
        unsigned int                         m_thread_count;
        unsigned int                         m_used_threads_count;
        // TODO: Provide our own fast linked-list underlying container implementation.
        std::queue<boost::function0<void> >  m_task_queue;
        boost::condition                     m_new_task_condition;
        boost::mutex                         m_new_task_mutex;        
        boost::mutex                         m_queue_mutex;
};

} // namespace miraclient

#endif // __MIRACLIENT_THREADPOOL_H
