/*
 * Client.cpp - by Alan Alvarez <alan.alvarez@us.army.mil>
 *
 * Mira GroupWare
 * Copyright (C) 2007-2008 Mira GroupWare Dev Team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <string>
#include <iostream>
#include <exception>
#include <cstring>
#include <boost/thread/mutex.hpp>

#include <boost/asio.hpp>

#include "Msg.h"
#include "AsioClient.h"
#include "Application.h"
#include "tools.h"

using boost::asio::ip::tcp;
using std::cout;
using std::endl;
using std::string;

namespace miraclient {

namespace network {

namespace asio {

AsioClient::AsioClient() : Client(), m_io(), m_socket(m_io), m_input_buffer_size(0), m_output_buffer_size(0)
{
    *m_input_buffer = '\0';
    m_input_buffer[INPUT_BUFFER_SIZE-1] = '\0';

    Application::set_client(dynamic_cast<Client*>(this));
}

void AsioClient::connect(const std::string& hostname, const unsigned int port, const std::string& username, const std::string& password)
{
    // Check if hostname is an IP address
    boost::system::error_code ec;
    boost::asio::ip::address::from_string(hostname, ec);

    boost::system::error_code l_error = boost::asio::error::host_not_found;

    if (ec)
    {
        std::string port_string = IntToString(port);
        tcp::resolver l_resolver(m_io);
        tcp::resolver::query l_query(hostname, IntToString(port));
        boost::system::error_code resolver_error;
        tcp::resolver::iterator Iter = l_resolver.resolve(l_query, resolver_error);
        if (!resolver_error)
        {
            tcp::resolver::iterator Iter_end;
            while (l_error && Iter != Iter_end)
            {
                std::cout << Iter->endpoint().address().to_string() << std::endl;
                m_socket.close();
                m_socket.connect(*Iter++, l_error);
            }
        }
        else
        {
            QMetaObject::invokeMethod(Application::get_gui_application()->m_mainWindow, "loginFailed", Qt::QueuedConnection, MIRA_QT_GENERIC_ARG(const QString&, QString(resolver_error.message().c_str())));
            return;
        }
    }
    else
    {
        m_socket.connect(boost::asio::ip::tcp::endpoint(boost::asio::ip::address::from_string(hostname), port), l_error);        
    }

    if (l_error)
    {
        QMetaObject::invokeMethod(Application::get_gui_application()->m_mainWindow, "loginFailed", Qt::QueuedConnection, MIRA_QT_GENERIC_ARG(const QString&, QString(l_error.message().c_str())));
        return;
    }
    else
    {
        m_remote_server_address = hostname;
        std::string login_string = "L ";
        login_string += username + " ";
        login_string += password;
        send(login_string);
    }

    receive();
}

void AsioClient::receive()
{
do
{
    boost::system::error_code l_error;

   size_t bytes_transferred = m_socket.read_some(boost::asio::buffer(get_input_buffer(), input_buffer_size()), l_error);

    if (l_error)
    {
        delete this;
        return;
    }

    char* current = m_input_buffer;

cout << "bytes_transferred: " << bytes_transferred << endl;

    m_input_buffer[(m_input_buffer_size+bytes_transferred)] = '\0'; // NULL-terminate the C-String

    bool digested = false;

    char* pos_found = NULL;
    while (pos_found = strchr(current, '\n'))
    {
        // Ideally messages should only have '\n' at the end, but check for (windows) "\r\n" just in case.
        // Also telnet clients (what I use for debugging) send "\r\n"
        // Put trailing NULL char so we can copy as cstring;
        if (*(pos_found-1) == '\r')
            *(pos_found-1) = '\0';
        else
            *pos_found = '\0';

        m_message_queue.push(current); // Push complete message.
        digested = true;
        cout << "IN: " << (m_message_queue.front()) << endl;
        if ( bytes_transferred > static_cast<size_t>( ((pos_found) - get_input_buffer())) ) // more than one complete message
        {
            current = pos_found+1;   // current is now right after '\n';
            continue;
        }
        else
        {
            current = NULL; // set current as null. This is so we know there aren't any more messages in the buffer.
            break;
        }
    }

    if (!current)
    {
        size_t bytes_copied = (bytes_transferred - (current - get_input_buffer()) );
        ::memmove(m_input_buffer, current, bytes_copied);
        m_input_buffer_size = bytes_copied;
    }
    else if (!digested)
    {
        m_input_buffer_size = bytes_transferred;
        // if buffer is full, flag error to user and empty buffer.
        if (input_buffer_size() == 0)
        {
            // send("E MSG TOO BIG");
            m_input_buffer_size = 0;
        }

        continue; // nothing more to do if message wasn't digested so start reading again.
    }
    else
    {
        m_input_buffer_size = 0;
    }


    if (digested)
    {
        while (!m_message_queue.empty())
        {
            // NOTE: the pointer returned by m_message_queue.front() NEEDS to be deleted in Msg::Parse().
            //       We have to store messages as pointer to make we still have valid a object after we pop()
            //       and the worker thread is accessing the object.
            Msg::Parse( m_message_queue.front() );
            m_message_queue.pop();
        }
    }
}
while (true);

}

void AsioClient::add_to_output_buffer(const char* msg, unsigned int size)
{
cout << "adding to buffer..." << endl;
    while (msg[size-1] == '\0')
        --size;

    // - If our buffer is not big enough to hold this, put as much as we can in buffer and send to client.
    // - Keep doing that until we have something that is small enough to not cause a buffer overflow.
    // - After that is done whatever remains will not be sent out.
    unsigned int pos = 0;
    for (;(m_output_buffer_size + size) > m_output_buffer_max_size;
        size -= m_output_buffer_max_size, pos += m_output_buffer_max_size)
    {
        boost::mutex::scoped_lock output_buffer_lock(m_output_buffer_mutex);
        m_output_buffer_size = m_output_buffer_max_size;
        std::memcpy(m_output_buffer, (msg+pos), m_output_buffer_size);
        output_buffer_lock.unlock();

        send();
    }

    if (size)
    {
        boost::mutex::scoped_lock output_buffer_lock(m_output_buffer_mutex);
        std::memcpy(m_output_buffer, (msg+pos), size);
        m_output_buffer_size += size;
        m_output_buffer[m_output_buffer_size]   = '\n'; // TODO: Maybe we shouldn't add this here, but let the user do it.
        output_buffer_lock.unlock();
    }
}

void AsioClient::send()
{
    boost::mutex::scoped_lock output_buffer_lock(m_output_buffer_mutex);
    boost::asio::write(m_socket, boost::asio::buffer(m_output_buffer, output_buffer_size()+1),
                       boost::asio::transfer_all());
    m_output_buffer_size = 0;
}

const size_t AsioClient::m_output_buffer_max_size = (OUTPUT_BUFFER_SIZE-2);

} // namespace asio

} // namespace network

} // namespace miraclient
