/*
 * Resource.h - by Alan Alvarez <alan.alvarez@us.army.mil>
 *
 * Mira GroupWare
 * Copyright (C) 2007-2009 Mira GroupWare Dev Team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __MIRASERVER_RESOURCE_H
#define __MIRASERVER_RESOURCE_H

#include <string>
// Microsoft Visual C++ doesn't provide a cstdint header.
#ifdef _MSC_VER // [
#include <boost/cstdint.hpp>
using boost::uint8_t;
#endif // _MSC_VER ]

class Resource
{
    public:
        static const uint8_t  READ_ACCESS       = 0x01;
        static const uint8_t  WRITE_ACCESS      = 0x10;
        static const uint8_t  READ_WRITE_ACCESS = 0x11;
        static const uint8_t  NO_ACCESS         = 0x00;

        Resource() : m_name(""), m_access_list()
        {}

        Resource(const std::string &name) : m_name(name), m_access_list()
        {}

        ~Resource()
        {}

        std::string get_name() const { return m_name; }

        void set_access(const unsigned int user, const uint8_t access)
        {
            m_access_list[user] = access;
        }
        void give_read_access(const unsigned int user)
        {
            AccessListType::iterator Iter = m_access_list.find(user);
            if (Iter != m_access_list.end())
                Iter->second = READ_ACCESS | Iter->second;
            else
                set_access(user, READ_ACCESS);
        }

        void give_write_access(const unsigned int user)
        {
            AccessListType::iterator Iter = m_access_list.find(user);
            if (Iter != m_access_list.end())
                Iter->second = WRITE_ACCESS | Iter->second;
            else
                set_access(user, WRITE_ACCESS);
        }

        void remove_read_access(const unsigned int user)
        {
            AccessListType::iterator Iter = m_access_list.find(user);
            if (Iter != m_access_list.end())
                Iter->second = WRITE_ACCESS & Iter->second;

            if (Iter->second == NO_ACCESS)
                m_access_list.erase(Iter);
        }

        void remove_write_access(const unsigned int user)
        {
            AccessListType::iterator Iter = m_access_list.find(user);
            if (Iter != m_access_list.end())
                Iter->second = READ_ACCESS & Iter->second;

            if (Iter->second == NO_ACCESS)
                m_access_list.erase(Iter);
        }

        uint8_t get_access(const unsigned int user)
        {
            AccessListType::iterator Iter = m_access_list.find(user);
            if (Iter != m_access_list.end())
                return Iter->second;
            else
                return NO_ACCESS;
        }

        bool has_read_access(const unsigned int user)
        {
            return bool(get_access(user) & READ_ACCESS);
        }

        bool has_write_access(const unsigned int user)
        {
            return bool(get_access(user) & WRITE_ACCESS);
        }

    private:
        friend class boost::serialization::access;
        typedef std::map<unsigned int, uint8_t> AccessListType;

        template<class Archive>
        void serialize(Archive & ar, const unsigned int /* file_version */)
        {
            ar & BOOST_SERIALIZATION_NVP(m_name);
            ar & BOOST_SERIALIZATION_NVP(m_access_list);
        }

        std::string     m_name;
        AccessListType  m_access_list;
};

#endif // __MIRASERVER__FIELD_H
