/*
 * PlainTextDirectory.h - by Alan Alvarez <alan.alvarez@us.army.mil>
 *
 * Mira GroupWare
 * Copyright (C) 2007-2008 Mira GroupWare Dev Team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __MIRASERVER__DIRECTORY_PLAINTEXTDIRECTORY_H

#include <iostream>
#include <string>
#include <fstream>
#include <ios>
#include <map>

#include <boost/serialization/nvp.hpp>
#include <boost/serialization/utility.hpp>
#include <boost/serialization/list.hpp>
#include <boost/serialization/version.hpp>

#include "Directory.h"
#include "User.h"
#include "WorkPlace.h"
#include "Resource.h"

namespace miraserver
{

namespace directory
{

class PlainTextDirectory : public Directory
{
    public:
        PlainTextDirectory(const std::string &file_name) : Directory(), m_filename(file_name),
                                                           m_user_list(), m_workplace_list()
        {  load(); }

        virtual ~PlainTextDirectory()
        {
            std::cout << "Saving from PlainTextDirectory destructor" << std::endl;
            save();
        }

        virtual unsigned int login(const std::string &username, const std::string &password);

        virtual bool add_user(User& user)
    	{
            // Take last id+1 in map if there are any items in it, if not use 1.
            if (m_user_list.size() > 0)
            {
                UserListType::iterator Iter = m_user_list.end();
                --Iter;
                user.set_id(Iter->first+1);
            }
            else
            {
                user.set_id(1);
            }

            bool ret_val = m_user_list.insert(UserListType::value_type(user.get_id(), user)).second;
            save();
            return ret_val;
    	}

        virtual bool add_workplace(WorkPlace& workplace) 
        { 
            // Take last id+1 in map if there are any items in it, if not use 1.
            if (m_workplace_list.size() > 0)
            {
                WorkPlaceListType::iterator Iter = m_workplace_list.end();
                --Iter;
                workplace.set_id(Iter->first+1);
            }
            else
            {
                workplace.set_id(1);
            }

            bool ret_val = m_workplace_list.insert(WorkPlaceListType::value_type(workplace.get_id(), workplace)).second;
            save();
            return ret_val;
        }

        virtual bool add_resource(const Resource& resource)
        {
            if (resource.get_name() != "")
            {
                bool ret_val = m_resource_list.insert(ResourceListType::value_type(resource.get_name(), resource)).second;
                save();
                return ret_val;
            }
            else
            {
                return false;
            }

        }

        virtual User find_user(const unsigned int user_id)
        {
            UserListType::iterator Iter = m_user_list.find(user_id);
            return Iter != m_user_list.end() ? Iter->second : User();
        }

        virtual User find_user(const std::string& username)
        {
            for (UserListType::iterator Iter = m_user_list.begin(); Iter != m_user_list.end(); ++Iter)
            {
                if (Iter->second.get_username() == username)
                    return Iter->second;
            }

            return User();
        }

        virtual WorkPlace find_workplace(const unsigned int workplace_id)
        {
            WorkPlaceListType::iterator Iter = m_workplace_list.find(workplace_id);
            return Iter != m_workplace_list.end() ? Iter->second : WorkPlace();
        }

        virtual WorkPlace find_workplace(const std::string&  workplace_name)
        {
            for (WorkPlaceListType::iterator Iter = m_workplace_list.begin(); Iter != m_workplace_list.end(); ++Iter)
            {
                if (Iter->second.get_workplace_name() == workplace_name)
                    return Iter->second;
            }

            return WorkPlace();
        }

        virtual Resource find_resource(const std::string& resource_name)
        {
            ResourceListType::iterator Iter = m_resource_list.find(resource_name);
            if (Iter != m_resource_list.end())
            {
                return Iter->second;
            }
            else
            {
                Resource("");
            }
        }

        virtual bool update_user(const User& user)
        {
            UserListType::iterator Iter = m_user_list.find(user.get_id());
            if (Iter != m_user_list.end())
            {
                Iter->second = user;
                save();
                return true;
            }

            return false;
        }

        virtual bool update_workplace(const WorkPlace& workplace)
        {
            WorkPlaceListType::iterator Iter = m_workplace_list.find(workplace.get_id());
            if (Iter != m_workplace_list.end())
            {
                Iter->second = workplace;
                save();
                return true;
            }

            return false;
        }

        virtual bool update_resource(const Resource& resource)
        {
            ResourceListType::iterator Iter = m_resource_list.find(resource.get_name());
            if (Iter != m_resource_list.end())
            {
                Iter->second = resource;
                save();
                return true;
            }
            else
            {
                return false;
            }
        }

        void save();
        void load();

    private:
        typedef std::map<unsigned int, User>      UserListType;
        typedef std::map<unsigned int, WorkPlace> WorkPlaceListType;
        typedef std::map<std::string, Resource>   ResourceListType;
        friend class boost::serialization::access;

        template<class Archive>
        void serialize(Archive & ar, const unsigned int /* file_version */)
        {
            ar & BOOST_SERIALIZATION_NVP(m_user_list);
            ar & BOOST_SERIALIZATION_NVP(m_workplace_list);
            ar & BOOST_SERIALIZATION_NVP(m_resource_list);
        }

        std::string m_filename;

        UserListType       m_user_list;
        WorkPlaceListType  m_workplace_list;
        ResourceListType   m_resource_list;
};

} // namespace directory

} // namespace miraserver

#endif // __MIRASERVER__DIRECTORY_PLAINTEXTDIRECTORY_H
