/*
 * TcpConnection.h - by Alan Alvarez <alan.alvarez@us.army.mil>
 *
 * Mira GroupWare
 * Copyright (C) 2007-2008 Mira GroupWare Dev Team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __MIRASERVER__NETWORK_TCPCONNECTION_H
#define __MIRASERVER__NETWORK_TCPCONNECTION_H

#include <string>
#include <iostream>
#include <queue>
#include <cstring>

#include <boost/thread.hpp>
#include <boost/thread/mutex.hpp>
#include <boost/thread/condition.hpp>

#include "ThreadPool.h"

#define INPUT_BUFFER_SIZE  1025 // 1 kilobyte (+1 for NULL char) for now.
#define OUTPUT_BUFFER_SIZE 1025 // 1 kilobyte (+1 '\n' char) for now.

namespace miraserver
{

namespace network
{

// A TcpConnection represents a connection to a client.
// It contains its own input/output buffer and its own socket.
// This class also contains and manages the thread pool statically.
class TcpConnection
{
    public:

        TcpConnection(void *socket) : m_user_id(0), m_input_buffer(), m_input_buffer_size(0), m_output_buffer(),
                                      m_output_buffer_size(0), m_socket(socket)
        {
            *m_input_buffer = '\0';
            m_input_buffer[INPUT_BUFFER_SIZE-1] = '\0';
        }

        ~TcpConnection();

        unsigned int get_user_id() { return m_user_id; }
        void set_user_id(const unsigned int user_id) { m_user_id = user_id; }

        void* get_socket() { return m_socket; }

        // NOTE: It is not necessary to pass a NULL character at the end of the string.
        //       Although doing so will not do any harm.
        void add_to_output_buffer(const char* msg, unsigned int size);

        void add_to_output_buffer(const char* msg)
        {
            add_to_output_buffer(msg, std::strlen(msg));
        }

        void add_to_output_buffer(const std::string& msg)
        {
            add_to_output_buffer(msg.c_str(), msg.length());
        }

        void send(const char* msg)
        {
            add_to_output_buffer(msg);
            std::cout << "[OUT]: " << msg << std::endl;
            send();
        }

        void send(const char* msg, unsigned int size)
        {
            add_to_output_buffer(msg, size);
            send();
        }

        void send(const std::string &msg)
        {
            add_to_output_buffer(msg);
            send();
        }

        void send()
        {
            boost::mutex::scoped_lock output_buffer_lock(m_output_buffer_mutex);
            send_to_client(*this);
            m_output_buffer_size = 0;
        }

        static void set_send_pointer(bool (*send_pointer)(TcpConnection& tcp_connection))
        {
            send_to_client = send_pointer;
        }

        static void set_receive_pointer(bool (*receive_pointer)(TcpConnection& tcp_connection))
        {
            receive_from_client = receive_pointer;
        }

        static void set_close_pointer(void (*close_pointer)(void* socket))
        {
            close_socket = close_pointer;
        }

        char* get_input_buffer()
        {
            return (m_input_buffer + m_input_buffer_size);
        }

        char* get_output_buffer()
        {
            return m_output_buffer;
        }

        size_t input_buffer_size() { return ((INPUT_BUFFER_SIZE-1) - m_input_buffer_size); } // -1 because we want to leave a space for NULL char.

        size_t output_buffer_size() { return m_output_buffer_size; }

        void handle_read(size_t bytes_transferred);

        bool is_logged_in() const { return (m_user_id != 0); } // User id of 0 means it's an invalid user (or not connected in this case)

        static void add_thread(const unsigned int threads_to_add = 1)
        {
            m_thread_pool.add_thread(threads_to_add);
        }

   private:

             static bool (*send_to_client)(TcpConnection& tcp_connection);
             static bool (*receive_from_client)(TcpConnection& tcp_connection);
             static void (*close_socket)(void* socket);

             unsigned int              m_user_id;  // User ID of the user using this TcpConnection.
             char                      m_input_buffer[INPUT_BUFFER_SIZE];
             size_t                    m_input_buffer_size;
             char                      m_output_buffer[OUTPUT_BUFFER_SIZE];
             size_t                    m_output_buffer_size;
static const size_t                    m_output_buffer_max_size;
             void*                     m_socket;
             std::queue<std::string*>  m_message_queue;
      static ThreadPool                m_thread_pool;
             boost::mutex              m_output_buffer_mutex;
};

} // namespace network

} // namespace miraserver

#endif // __MIRASERVER__NETWORK_TCPCONNECTION_H
