/*
 * TcpConnection.cpp - by Alan Alvarez <alan.alvarez@us.army.mil>
 *
 * Mira GroupWare
 * Copyright (C) 2007-2008 Mira GroupWare Dev Team
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <iostream>
#include <cstring>
#include <string>

#include "ThreadPool.h"
#include "TcpConnection.h"
#include "Application.h"
#include "Msg.h"
#include "field_defs.h"
#include "tools.h"

using std::string;

// for debugging.
using std::cout;
using std::endl;

namespace miraserver
{

namespace network
{

TcpConnection::~TcpConnection()
{
    if(is_logged_in())
        Application::get_server().remove_connection(m_user_id);

    // delete peding messages in queue.
    // TODO: We might actually want to process these messages before deleting this object
    unsigned int size = m_message_queue.size();
    for (unsigned int i = 0;  i < size; ++i, m_message_queue.pop())
        delete m_message_queue.front();

    cout << "dropping connection" << endl;

    close_socket(m_socket);
}

void TcpConnection::add_to_output_buffer(const char* msg, unsigned int size)
{
    while (msg[size-1] == '\0')
        --size;

    // - If our buffer is not big enough to hold this, put as much as we can in buffer and send to client.
    // - Keep doing that until we have something that is small enough to not cause a buffer overflow.
    // - After that is done whatever remains will not be sent out.
    unsigned int pos = 0;
    for (;(m_output_buffer_size + size) > m_output_buffer_max_size;
        size -= m_output_buffer_max_size, pos += m_output_buffer_max_size)
    {
        boost::mutex::scoped_lock output_buffer_lock(m_output_buffer_mutex);
        m_output_buffer_size = m_output_buffer_max_size;
        std::memcpy(get_output_buffer(), (msg+pos), m_output_buffer_size);
        output_buffer_lock.unlock();

        send();
    }

    if (size)
    {
        boost::mutex::scoped_lock output_buffer_lock(m_output_buffer_mutex);
        std::memcpy(get_output_buffer(), (msg+pos), size);
        m_output_buffer_size += size;
        m_output_buffer[m_output_buffer_size]   = '\n';
        output_buffer_lock.unlock();
    }
}

void TcpConnection::handle_read(size_t bytes_transferred)
{
    if (!bytes_transferred)
    {
        delete this;
        return;
    }

    char* current = m_input_buffer;

    cout << "bytes_transferred: " << bytes_transferred << endl;

    m_input_buffer[(m_input_buffer_size+bytes_transferred)] = '\0'; // NULL-terminate the C-String

    bool digested = false;

    char* pos_found = NULL;
    while (pos_found = strchr(current, '\n'))
    {
        // Ideally messages should only have '\n' at the end, but check for (windows) "\r\n" just in case.
        // Also telnet clients (what I use for debugging) send "\r\n"
        // Put trailing NULL char so we can copy as cstring;
        if (*(pos_found-1) == '\r')
            *(pos_found-1) = '\0';
        else
            *pos_found = '\0';

        m_message_queue.push(new string(current)); // Push complete message.
        digested = true;
        cout << "IN: " << (*m_message_queue.front()) << endl;
        if ( bytes_transferred > static_cast<size_t>( ((pos_found) - get_input_buffer())) ) // more than one complete message
        {
            current = pos_found+1;   // current is now right after '\n';
            continue;
        }
        else
        {
            current = NULL; // set current as null. This is so we know there aren't any more messages in the buffer.
            break;
        }
    }

    if (!current)
    {
        size_t bytes_copied = (bytes_transferred - (current - get_input_buffer()) );
        ::memmove(m_input_buffer, current, bytes_copied);
        m_input_buffer_size = bytes_copied;
    }
    else if (!digested)
    {
        m_input_buffer_size = bytes_transferred;

        // if buffer is full, flag error to user and empty buffer.
        if (input_buffer_size() == 0)
        {
            send("E MSG TOO BIG");
            m_input_buffer_size = 0;
        }

        receive_from_client(*this); // keep reading from client.
        return; // nothing more to do if message wasn't digested so return.
    }
    else
    {
        m_input_buffer_size = 0;
    }
      size_t token_pos = (*m_message_queue.front()).find(" ");
      string token = (*m_message_queue.front()).substr(0, token_pos);

    // User needs to login before doing anything else
    //Except, for now, NU(new user) don't have to login for creating new account.
    if((token.compare("NU") != 0) && (token.compare("L") != 0) && (!is_logged_in()))
    {
        send("E Need To Login");
        delete this;
        return;
    }

    if(digested)
    {
        while(!m_message_queue.empty())
        {
            // NOTE: the pointer returned by m_message_queue.front() NEEDS to be deleted in Msg::Parse().
            // We have to store messages as pointer to make we still have valid a object after we pop()
            //and the worker thread is accessing the object.

            TcpConnection::m_thread_pool.add_task( boost::bind(&Msg::Parse, this, m_message_queue.front()) );

            m_message_queue.pop();
        }
    }
    receive_from_client(*this); // keep reading from client.
}

bool (*TcpConnection::send_to_client)(TcpConnection& tcp_connection) = NULL;
bool (*TcpConnection::receive_from_client)(TcpConnection& tcp_connection) = NULL;
void (*TcpConnection::close_socket)(void* socket) = NULL;
ThreadPool TcpConnection::m_thread_pool = ThreadPool();
const size_t TcpConnection::m_output_buffer_max_size = (OUTPUT_BUFFER_SIZE-2);

} // namespace network

} // namespace miraserver

